// "Direct Manipulation of Procedural Implicit Surfaces", SIGGRAPH Asia 2024
// MARZIA RISO, Sapienza University of Rome, Italy and Adobe, France
// ÉLIE MICHEL, Adobe, France
// AXEL PARIS, Adobe, France
// VALENTIN DESCHAINTRE, Adobe, United Kingdom
// MATHIEU GAILLARD, Adobe, USA
// FABIO PELLACINI, University of Modena and Reggio Emilia, Italy

// -------------------------------------------------
// Preamble
// -------------------------------------------------

// Output of an augmented implicit function
struct AugmentedOutput {
	float sdf;
	vec4 coparam; // NB: The last dimension is the integer path index
};

// Utility functions that returns a position relative to a box/square
// if outside, and the interior distance if inside.
vec4 boxFrame(in vec3 pos, in vec3 dims) {
	vec3 h = abs(pos) - dims;
	return vec4(
		max(h, vec3(0.0)),
		min(0.0, max(h.x, max(h.y, h.z)))
	);
}

vec3 squareFrame(in vec2 pos, in vec2 dims) {
	vec2 h = abs(pos) - dims;
	return vec3(
		max(h, vec2(0.0)),
		min(0.0, max(h.x, h.y))
	);
}

mat2 rotationMatrix(float angle) {
	float c = cos(angle);
	float s = sin(angle);
	return mat2(c, s, -s, c);
}

float ndot(in vec2 a, in vec2 b){
	float x = a.x * b.x;
	float y = a.y * b.y;
	return x - y;
}


// -------------------------------------------------
// Primitives
// -------------------------------------------------

AugmentedOutput evalPrimSphere(in vec3 pos, float radius) {
	float sdf = length(pos) - radius;
	vec4 coparam = vec4(pos / radius, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimBox(in vec3 pos, in vec3 dims) {
	vec4 box = boxFrame(pos, dims);
	float sdf = length(box.xyz) + box.w;
	vec4 coparam = vec4(pos / dims, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimPlane(in vec3 pos, in vec3 normal, float offset) {
	float sdf = dot(normalize(normal), pos) - offset;
	vec4 coparam = vec4(pos - offset * normal, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimCylinder(in vec3 pos, float height, float radius) {
	vec3 sq = squareFrame(vec2(length(pos.xz), pos.y), vec2(radius, height));
	float sdf = length(sq.xy) + sq.z;
	vec4 coparam = vec4(pos.x / radius, pos.y / height, pos.z / radius, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimEllipsoid(in vec3 pos, in vec3 radii) {
	float l = length(pos / radii);
	float sdf = (l - 1.0) * l / length(pos / (radii * radii));
	vec4 coparam = vec4(pos / radii, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimCone(in vec3 pos, float height, float radius) {
	vec2 h = vec2(length(pos.xz), pos.y - height);
	vec2 g = vec2(radius, -height);
	vec2 u = h - g * clamp(dot(g, h) / dot(g, g), 0.0, 1.0);
	vec2 v = h - g * vec2(clamp(h.x / g.x, 0.0, 1.0), 1.0);
	float s = sign(max((h.y - g.y) * sign(g.y), (g.y * h.x - g.x * h.y) * sign(g.y)));
	float sdf = s * sqrt(min(dot(u, u), dot(v, v)));
	vec4 coparam = vec4(pos.x / radius, pos.y / height, pos.z / radius, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimCapsule(in vec3 pos, vec3 start, vec3 end, float radius) {
	// Y-aligned capsule
	vec3 pa = pos - start;
	vec3 ba = end - start;
	float h = clamp(dot(pa, ba) / dot(ba, ba), 0.0, 1.0) - radius;
	float sdf = length(pa - ba * h);

	vec3 ax = normalize(cross(ba, vec3(0.0, 0.0, 1.0)));
	vec3 ay = normalize(cross(normalize(cross(ba, vec3(0.0, 0.0, 1.0))), pa));
	vec3 az = normalize(ba);
	mat3 m = mat3(ax, ay, az);

	vec3 tpos = m * pa + vec3(0.0, radius, 0.0);
	vec3 ty = m * (pos - start) + vec3(0.0, radius, 0.0);
	vec4 coparam = vec4(
		tpos.x / radius, 
		mix(-1.0, 1.0, ty.y) / length(ba + 2.0 * radius),
		tpos.z / radius, 0.0);

	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimRhombus(in vec3 pos, float width, float depth, float height, float radius) {
	vec3 abs_pos = abs(pos);
	vec2 b = vec2(width, depth);
	float f = clamp(ndot(b, b - 2.0 * abs_pos.xz) / dot(b, b), -1.0, 1.0);
	vec2 q = vec2(length(abs_pos.xz - 0.5 * b * vec2(1.0 - f, 1.0 + f)) * 
			sign(abs_pos.x * b.y + abs_pos.z * b.x - b.x * b.y) - width, abs_pos.y - height);
	float sdf = min(max(q.x, q.y), 0.0) + length(max(q, 0.0));	

	vec4 coparam = vec4(
		pos.x / (width + radius),
		pos.y / height,
		pos.z / (depth + radius),
		0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimTorus(in vec3 pos, float radius_major, float radius_minor) {
	vec2 q = vec2(length(pos.xz) - radius_major, pos.y);
	float sdf = length(q) - radius_minor;
	vec4 coparam = vec4(
		atan(pos.y, length(pos.xy) - radius_major),
		atan(pos.z, pos.x),
		0.0, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimHexagonalPrism(in vec3 pos, float radius, float height) {
	const vec3 k = vec3(-0.8660254, 0.5, 0.57735);
	vec3 abs_pos = abs(pos);
	vec2 t = pos.xy - 2.0 * min(dot(k.xy, abs_pos.xy), 0.0) * k.xy;
 	vec2 d = vec2(length(t.xy - vec2(clamp(t.x, -k.z * height, k.z * radius), radius)) * sign(t.y - radius), abs_pos.z - height);
	float sdf = min(max(d.x, d.y), 0.0) + length(max(d, 0.0));

	vec4 coparam = vec4(pos.x / radius, pos.y / height, pos.z / radius, 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimElongatedCylinder(in vec3 pos, float height, float radius, float elong_width, float elong_height) {
	vec4 box = boxFrame(pos, vec3(elong_width, 0.0, elong_height));
	vec3 sq = squareFrame(vec2(length(box.xz), box.y), vec2(radius, height));
	float sdf = box.w + length(sq.xy) + sq.z;
	vec4 coparam = vec4(pos / vec3(elong_width + radius, height, elong_height + radius), 0.0);
	return AugmentedOutput(sdf, coparam);
}

AugmentedOutput evalPrimElongatedEllipsoid(in vec3 pos, in vec3 radii, float elong_width, float elong_depth) {
	vec4 box = boxFrame(pos, vec3(elong_width, 0.0, elong_depth));
	float l = length(box.xyz / radii);
	float sdf = box.w + (l - 1.0) * l / length(box.xyz / (radii * radii));
	vec4 coparam = vec4(pos / (vec3(elong_width, 0.0, elong_depth) + radii), 0.0);
	return AugmentedOutput(sdf, coparam);
}

// -------------------------------------------------
// Unary Operators
// -------------------------------------------------

vec3 preOpTranslate(in vec3 pos, in vec3 offset) {
	return pos - offset;
}
AugmentedOutput postOpTranslate(in AugmentedOutput from_input, in vec3 offset) {
	return from_input;
}

vec3 preOpRotate(in vec3 pos, in mat3 rotation) {
	return rotation * pos;
}
AugmentedOutput postOpRotate(in AugmentedOutput from_input, in mat3 rotation) {
	return from_input;
}

vec3 preOpScale(in vec3 pos, float scale) {
	return pos / scale;
}
AugmentedOutput postOpScale(in AugmentedOutput from_input, float scale) {
	float sdf = from_input.sdf * scale;
	return AugmentedOutput(sdf, from_input.coparam);
}

vec3 preOpBend(in vec3 pos, float angle, float start, float end) {
	float fac = smoothstep(0.0, 1.0, clamp((pos.z - start) / (end - start), 0.0, 1.0));
	return vec3(rotationMatrix(angle * fac) * pos.xy, pos.z);
}
AugmentedOutput postOpBend(in AugmentedOutput from_input, float angle, float start, float end){
	return from_input;
}

vec3 preOpTwist(in vec3 pos, float angle, float start, float end) {
	float fac = smoothstep(0.0, 1.0, clamp((pos.y - start) / (end - start), 0.0, 1.0));
	vec2 twisted = rotationMatrix(angle * fac) * pos.xz;
	return vec3(twisted.x, pos.y, twisted.y);

}

AugmentedOutput postOpTwist(in AugmentedOutput from_input, float angle, float start, float end) {
	return from_input;
}

vec3 preOpDisplace(in vec3 pos, float mask, in vec3 intensity) {
	return pos - pos * intensity * mask;
}
AugmentedOutput postOpDisplace(in AugmentedOutput from_input, float mask, in vec3 intensity) {
	return from_input;
}

vec3 preOpMirrorX(in vec3 pos) {
	return vec3(-pos.x, pos.yz);
}
AugmentedOutput postOpMirrorX(in AugmentedOutput from_input) {
	return from_input;
}

vec3 preOpMirrorY(in vec3 pos) {
	return vec3(pos.x, -pos.y, pos.z);
}
AugmentedOutput postOpMirrorY(in AugmentedOutput from_input) {
	return from_input;
}

vec3 preOpMirrorZ(in vec3 pos) {
	return vec3(pos.xy, -pos.z);
}
AugmentedOutput postOpMirrorZ(in AugmentedOutput from_input) {
	return from_input;
}

vec3 preOpRounding(in vec3 pos, float radius) {
	return pos;
}
AugmentedOutput postOpRounding(in AugmentedOutput from_input, float radius) {
	float sdf = from_input.sdf - radius;
	return AugmentedOutput(sdf, from_input.coparam);
}

vec3 preOpShell(in vec3 pos, float thickness) {
	return pos;
}
AugmentedOutput postOpShell(in AugmentedOutput from_input, float thickness) {
	float sdf = abs(from_input.sdf) - thickness;
	return AugmentedOutput(sdf, from_input.coparam);
}

// -------------------------------------------------
// Binary Operators
// -------------------------------------------------

vec3[2] preOpUnion(in vec3 pos) {
	return vec3[2](pos, pos);
}
AugmentedOutput postOpUnion(in AugmentedOutput from_inputs[2], int pid_offset) {
	float sdf = min(from_inputs[0].sdf, from_inputs[1].sdf);
	vec4 coparam = mix(
		from_inputs[0].coparam,
		from_inputs[1].coparam + vec4(0.0, 0.0, 0.0, float(pid_offset)),
		step(from_inputs[1].sdf, from_inputs[0].sdf)
	);
	return AugmentedOutput(sdf, coparam);
}

vec3[2] preOpIntersection(in vec3 pos) {
	return vec3[2](pos, pos);
}
AugmentedOutput postOpIntersection(in AugmentedOutput from_inputs[2], int pid_offset) {
	float sdf = max(from_inputs[0].sdf, from_inputs[1].sdf);
	vec4 coparam = mix(
		from_inputs[0].coparam,
		from_inputs[1].coparam + vec4(0.0, 0.0, 0.0, float(pid_offset)),
		step(from_inputs[0].sdf, from_inputs[1].sdf)
	);
	return AugmentedOutput(sdf, coparam);
}

vec3[2] preOpDifference(in vec3 pos) {
	return vec3[2](pos, pos);
}
AugmentedOutput postOpDifference(in AugmentedOutput from_inputs[2], int pid_offset) {
	float sdf = max(from_inputs[0].sdf, -from_inputs[1].sdf);
	vec4 coparam = mix(
		from_inputs[0].coparam,
		from_inputs[1].coparam + vec4(0.0, 0.0, 0.0, float(pid_offset)),
		step(from_inputs[0].sdf, -from_inputs[1].sdf)
	);
	return AugmentedOutput(sdf, coparam);
}

vec3[2] preOpSmoothUnion(in vec3 pos, float smooth_factor) {
	return vec3[2](pos, pos);
}
AugmentedOutput postOpSmoothUnion(in AugmentedOutput from_inputs[2], float smooth_factor, int pid_offset) {
	float h = max(smooth_factor - abs(from_inputs[0].sdf - from_inputs[1].sdf), 0.0) / smooth_factor;
	float sdf = min(from_inputs[0].sdf, from_inputs[1].sdf) - h * h * smooth_factor * 1.0/4.0;
	vec4 coparam = mix(
		from_inputs[0].coparam, 
		from_inputs[1].coparam + vec4(0.0, 0.0, 0.0, pid_offset),
		step(from_inputs[1].sdf, from_inputs[0].sdf));
	return AugmentedOutput(sdf, coparam);
}

vec3[2] preOpSmoothIntersection(in vec3 pos, float smooth_factor) {
	return vec3[2](pos, pos);
}
AugmentedOutput postOpSmoothIntersection(in AugmentedOutput from_inputs[2], float smooth_factor, int pid_offset) {
	float h = max(smooth_factor - abs(from_inputs[0].sdf - from_inputs[1].sdf), 0.0) / smooth_factor;
	float sdf = max(from_inputs[0].sdf, from_inputs[1].sdf) + h * h * smooth_factor * 1.0/4.0;

	vec4 coparam = mix(
		from_inputs[0].coparam, 
		from_inputs[1].coparam + vec4(0.0, 0.0, 0.0, pid_offset),
		step(from_inputs[0].sdf, from_inputs[1].sdf));
	return AugmentedOutput(sdf, coparam);
}

vec3[2] preOpSmoothDifference(in vec3 pos, float smooth_factor) {
	return vec3[2](pos, pos);
}
AugmentedOutput postOpSmoothDifference(in AugmentedOutput from_inputs[2], float smooth_factor, int pid_offset) {
	float h = max(smooth_factor - abs(-from_inputs[1].sdf - from_inputs[0].sdf), 0.0) / smooth_factor;
	float sdf = max(from_inputs[0].sdf, -from_inputs[1].sdf) + h * h * smooth_factor * 1.0/4.0;

	vec4 coparam = mix(
		from_inputs[0].coparam, 
		from_inputs[1].coparam + vec4(0.0, 0.0, 0.0, pid_offset),
		step(from_inputs[0].sdf, -from_inputs[1].sdf));
	return AugmentedOutput(sdf, coparam);
}